#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import io
import sys
import time
import string
import atexit
import threading

from pathlib import Path
from typing import Optional

from lib_not_dr.loggers import LogLevel
from lib_not_dr.types.options import Options
from lib_not_dr.loggers.structure import LogMessage
from lib_not_dr.loggers.formatter import BaseFormatter, StdFormatter

# fmt: off
__all__ = [
    "BaseOutputStream",
    "StdioOutputStream",
    "FileCacheOutputStream"
]
# fmt: on


class BaseOutputStream(Options):
    name = "BaseOutputStream"

    level: int = LogLevel.info
    enable: bool = True

    formatter: BaseFormatter

    def write_stdout(self, message: LogMessage) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.write_stdout is not implemented")

    def write_stderr(self, message: LogMessage) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.write_stderr is not implemented")

    def flush(self) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.flush is not implemented")

    def close(self) -> None:
        self.enable = False


class StdioOutputStream(BaseOutputStream):
    name = "StdioOutputStream"

    level: int = LogLevel.info
    formatter: BaseFormatter = StdFormatter()
    use_stderr: bool = True

    def write_stdout(self, message: LogMessage) -> None:
        if not self.enable:
            return None
        if message.level < self.level:
            return None
        out_msg = self.formatter.format_message(message)
        if message.flush is not None:
            print(out_msg, end="", flush=message.flush)
        else:
            print(out_msg, end="", flush=True)
        return None

    def write_stderr(self, message: LogMessage) -> None:
        if not self.enable:
            return None
        if message.level < self.level:
            return None
        if self.use_stderr:
            if message.flush is not None:
                print(
                    self.formatter.format_message(message),
                    end="",
                    flush=message.flush,
                    file=sys.stderr,
                )
            else:
                print(
                    self.formatter.format_message(message),
                    end="",
                    flush=True,
                    file=sys.stderr,
                )
        else:
            self.write_stdout(message)
        return None

    def flush(self) -> None:
        """
        flush stdout and stderr
        :return: None
        """
        print("", end="", flush=True)
        print("", end="", flush=True, file=sys.stderr)
        return None


class FileCacheOutputStream(BaseOutputStream):
    name = "FileCacheOutputStream"

    level: int = LogLevel.info
    formatter: BaseFormatter = StdFormatter(enable_color=False)
    text_cache: io.StringIO = None  # type: ignore

    flush_counter: int = 0
    # 默认 10 次 flush 一次
    flush_count_limit: int = 10
    flush_time_limit: int = 10  # time limit in sec, 0 means no limit
    flush_timer: Optional[threading.Timer] = None

    file_path: Path = Path("./logs")
    # if contain {time} -> time.strftime("%Y-%m-%d_%H-%M-%S", time.gmtime(time.time)
    file_name: str
    # file mode: always 'a'
    file_encoding: str = "utf-8"
    # do file swap or not
    file_swap: bool = False
    at_exit_register: bool = False

    file_swap_counter: int = 0
    file_swap_name_template: str = "${name}-${counter}.log"
    # ${name} -> file_name
    # ${counter} -> file_swap_counter
    # ${log_time} -> time when file swap ( round(time.time()) )
    # ${start_time} -> start time of output stream ( round(time.time()) )
    current_file_name: Optional[str] = None
    file_start_time: Optional[int] = None

    # log file swap triggers
    # 0 -> no limit
    file_size_limit: int = 0  # size limit in kb
    file_time_limit: int = 0  # time limit in sec 0
    file_swap_on_both: bool = False  # swap file when both size and time limit reached

    def init(self, **kwargs) -> bool:
        # 时间取整
        self.file_start_time = round(time.time())
        # 初始化文件名
        if "{time}" in self.file_name:
            self.file_name = self.file_name.format(time=time.strftime(
                "%Y-%m-%d_%H-%M-%S", time.gmtime(self.file_start_time)
            ))
        # 初始化缓存
        if self.text_cache is None:
            self.text_cache = io.StringIO()
        # 检查文件名
        self.get_file_path()
        return False

    def _write(self, message: LogMessage) -> None:
        """
        write message to text cache
        默认已经检查过了
        :param message: message to write
        :return: None
        """
        self.text_cache.write(self.formatter.format_message(message))
        self.flush_counter += 1
        if message.flush or self.flush_counter >= self.flush_count_limit:
            self.flush()
        else:
            if self.flush_time_limit > 0:
                if self.flush_timer is None or not self.flush_timer.is_alive():
                    self.flush_timer = threading.Timer(self.flush_time_limit, self.flush)
                    self.flush_timer.daemon = True
                    self.flush_timer.start()
            if not self.at_exit_register:
                atexit.register(self.flush)
                self.at_exit_register = True
        return None

    def write_stdout(self, message: LogMessage) -> None:
        if not self.enable:
            return None
        if message.level < self.level:
            return None
        self._write(message)
        return None

    def write_stderr(self, message: LogMessage) -> None:
        if not self.enable:
            return None
        if message.level < self.level:
            return None
        self._write(message)
        return None

    def get_file_path(self) -> Path:
        """
        get file path
        :return:
        """
        if (current_file := self.current_file_name) is None:
            if not self.file_swap:
                # 直接根据 file name 生成文件
                current_file = Path(self.file_path) / self.file_name
                self.current_file_name = str(current_file)
                return current_file
            template = string.Template(self.file_swap_name_template)
            file_name = template.safe_substitute(
                name=self.file_name,
                counter=self.file_swap_counter,
                log_time=round(time.time()),
                start_time=self.file_start_time,
            )
            current_file = Path(self.file_path) / file_name
            self.current_file_name = str(current_file)
        else:
            current_file = Path(current_file)
        return current_file

    def check_flush(self) -> Path:
        current_file = self.get_file_path()
        # 获取当前文件的路径
        if not self.file_swap:
            # 不需要 swap 直接返回
            return current_file
        # 检查是否需要 swap
        size_pass = True
        if self.file_size_limit > 0:
            file_size = current_file.stat().st_size / 1024  # kb
            if file_size > self.file_size_limit:  # kb
                size_pass = False
        time_pass = True
        if self.file_time_limit > 0:
            file_time = round(time.time()) - current_file.stat().st_mtime
            if file_time > self.file_time_limit:
                time_pass = False
        if (self.file_swap_on_both and size_pass and time_pass) or (
                not self.file_swap_on_both and (size_pass or time_pass)):
            # 两个都满足
            # 或者只有一个满足
            if size_pass and time_pass:
                self.file_swap_counter += 1
                # 生成新的文件名
                return self.get_file_path()
        return current_file

    def flush(self) -> None:
        new_cache = io.StringIO()  # 创建新的缓存
        self.flush_counter = 0  # atomic, no lock
        old_cache, self.text_cache = self.text_cache, new_cache
        text = old_cache.getvalue()
        old_cache.close()  # 关闭旧的缓存
        if text == "":
            return None
        current_file = self.check_flush()
        # 检查文件是否存在
        if not current_file.exists():
            current_file.parent.mkdir(parents=True, exist_ok=True)
            current_file.touch(exist_ok=True)
        with current_file.open("a", encoding=self.file_encoding) as f:
            f.write(text)
        return None

    def close(self) -> None:
        super().close()
        self.flush()
        self.text_cache.close()
        atexit.unregister(self.flush)
        return None



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import time
import inspect
from types import FrameType
from typing import List, Optional, Union

from lib_not_dr.loggers import LogLevel
from lib_not_dr.types.options import Options
from lib_not_dr.loggers.structure import LogMessage
from lib_not_dr.loggers.outstream import BaseOutputStream, StdioOutputStream


class Logger(Options):
    name = "Logger-v2"

    outputs: List[BaseOutputStream] = [StdioOutputStream()]

    logger_name: str = "root"
    default_tag: Optional[str] = None

    enable: bool = True
    level: int = 20  # info

    def clone_logger(self) -> "Logger":
        """
        Clone a new loggers with the same config.

        Returns:
            Logger: The cloned loggers.
        """
        return Logger(
            logger_name=self.logger_name,
            enable=self.enable,
            level=self.level,
            outputs=self.outputs.copy(),
            default_tag=self.default_tag,
        )

    def log_for(self, level: int) -> bool:
        """
        Check if logging is enabled for a specific level.

        Args:
            level (int): The logging level to check.

        Returns:
            bool: True if logging is enabled for the given level, False otherwise.
        """
        return self.enable and level >= self.level

    def add_output(self, output: BaseOutputStream) -> None:
        """
        Add an output to the list of outputs.
        :param output:
        :return:
        """
        self.outputs.append(output)
        self.level = min(self.level, output.level)

    def remove_output(self, output: BaseOutputStream) -> None:
        """
        Remove an output from the list of outputs.
        :param output: BaseOutputStream
        :return:
        """
        self.outputs.remove(output)
        self.level = max(self.level, *[output.level for output in self.outputs])

    @property
    def tag(self):
        """
        Get the default tag.
        :return:
        """
        return self.default_tag

    @tag.setter
    def tag(self, tag: str) -> None:
        """
        Set the default tag.
        :param tag: str
        :return: None
        """
        self.default_tag = tag

    def set_tag(self, tag: str) -> "Logger":
        """
        Set the default tag.
        :param tag: str
        :return: None
        """
        self.default_tag = tag
        return self

    def clear_tag(self) -> None:
        """
        Clear the default tag.
        :return:
        """
        self.default_tag = None

    @property
    def global_level(self) -> int:
        """
        Get the global logging level.

        Returns:
            int: The global logging level.
        """
        return self.level

    @global_level.setter
    def global_level(self, level: int) -> None:
        """
        Set the global logging level.

        Args:
            level (int): The global logging level.

        Returns:
            None
        """
        self.level = level
        for output in self.outputs:
            output.level = level

    def make_log(
            self,
            messages: Union[list, tuple],
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            level: int = 20,  # info
            # log_time: Optional[float] = None,
            # logger_name: str = 'root',
            # logger_tag: Optional[str] = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        # 检查是否需要记录
        if not self.log_for(level):
            return
        log_time = time.time_ns()
        # 处理堆栈信息
        if stack_trace is None:
            # 尝试获取堆栈信息
            if (stack := inspect.currentframe()) is not None:
                # 如果可能 尝试获取上两层的堆栈信息
                if (up_stack := stack.f_back) is not None:
                    if (upper_stack := up_stack.f_back) is not None:
                        stack_trace = upper_stack
                    else:
                        stack_trace = up_stack
                else:
                    stack_trace = stack
        # 处理标签
        if tag is None and self.default_tag is not None:
            tag = self.default_tag

        message = LogMessage(
            messages=messages,  # type: ignore
            end=end,
            split=split,
            flush=flush,
            level=level,
            log_time=log_time,
            logger_name=self.logger_name,
            logger_tag=tag,
            stack_trace=stack_trace,
        )
        if level >= 30:  # WARN
            for output in self.outputs:
                output.write_stderr(message)
        else:
            for output in self.outputs:
                output.write_stdout(message)
        # done?
        # 20231106 00:06

    def info(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.info):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.info,
            stack_trace=stack_trace,
        )

    def trace(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.trace):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.trace,
            stack_trace=stack_trace,
        )

    def fine(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.fine):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.fine,
            stack_trace=stack_trace,
        )

    def debug(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.debug):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.debug,
            stack_trace=stack_trace,
        )

    def warn(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.warn):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.warn,
            stack_trace=stack_trace,
        )

    def error(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.error):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.error,
            stack_trace=stack_trace,
        )

    def fatal(
            self,
            *message,
            tag: Optional[str] = None,
            end: str = "\n",
            split: str = " ",
            flush: bool = None,
            stack_trace: Optional[FrameType] = None,
    ) -> None:
        if not self.log_for(LogLevel.fatal):
            return
        self.make_log(
            messages=message,
            tag=tag,
            end=end,
            split=split,
            flush=flush,
            level=LogLevel.fatal,
            stack_trace=stack_trace,
        )



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from typing import TYPE_CHECKING, Dict, Union, List, Callable

if TYPE_CHECKING:
    from lib_not_dr.nuitka import reader, compile

nuitka_config_type = Dict[str, Union[str, bool, List[Union[str, tuple]]]]
raw_config_type = Dict[str, Union[nuitka_config_type, str]]
parse_config_function = Callable[[raw_config_type], nuitka_config_type]

# fmt: off
__all__ = [
    "reader",
    "compile",
    "raw_config_type",
    "parse_config_function",
    "nuitka_config_type"
]
# fmt: on




#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import shutil
import traceback
from io import StringIO
from typing import (
    get_type_hints,
    Type,
    List,
    Union,
    Dict,
    Any,
    Callable,
    Tuple,
    Optional,
    TYPE_CHECKING,
    Iterable,
)

# fmt: off
__all__ = [
    "get_type_hints_",
    "Options",
    "OptionsError",
    "OptionNotFound",
    "OptionNotFilled",
    "OptionNameNotDefined",
]
# fmt: on


def get_type_hints_(cls: Type):
    try:
        return get_type_hints(cls)
    except ValueError:
        return get_type_hints(cls, globalns={})


def to_str_value_(value: Any) -> Any:
    """递归的将输入值的每一个非 builtin type 转换成 str"""
    if isinstance(value, (str, bytes, bytearray, int, float, bool, type(None))):
        return value
    elif isinstance(value, dict):
        return {k: to_str_value_(v) for k, v in value.items()}
    elif isinstance(value, (list, Iterable)):
        return [to_str_value_(v) for v in value]
    else:
        return str(value)


class OptionsError(Exception):
    """option 的错误基类"""


class OptionNameNotDefined(OptionsError):
    """向初始化的 option 里添加了一个不存在于选项里的选项"""


class OptionNotFound(OptionsError):
    """某个选项没有找到"""


class OptionNotFilled(OptionsError):
    """某个选项没有填写"""


class Options:
    """
    一个用于存储选项 / 提供 API 定义 的类
    用法:
    存储配置: 继承 Options 类
            在类里定义 option: typing
            (可选 定义 name: str = 'Option Base' 用于在打印的时候显示名字)
    提供 API 接口: 继承 Options 类
            在类里定义 option: typing
            定义 一些需要的方法
            子类: 继承 新的 Options 类
                实现定义的方法
    _check_options: 用于检查是否输入的选项都是已经定义的选项
    _check_filled: 用于检查是否所有的选项都已经填写
    """

    name = "Option Base"
    cached_options: Dict[str, Union[str, Any]] = {}

    # 用于检查是否输入的选项都是已经定义的选项
    _check_options: bool = True
    # 用于检查是否所有的选项都已经填写
    _check_filled: bool = False

    def __init__(self, **kwargs):
        """
        创建一个新的 Options 的时候的配置
        如果存在 init 方法 会在设置完 kwargs 之后运行子类的 init 方法
        :param kwargs: 需要设置的选项
        """
        if TYPE_CHECKING:
            self._options: Dict[str, Union[Callable, object]] = {}
        self.flush_option()
        miss_list = []
        if self._check_filled:
            miss_list = [key for key in self.cached_options]
        for option, value in kwargs.items():
            if option not in self.cached_options and self._check_options:
                raise OptionNameNotDefined(
                    f"option: {option} with value: {value} is not defined"
                )
            setattr(self, option, value)
            if self._check_filled:
                miss_list.remove(option)
        run_load_file = True
        if hasattr(self, "init"):
            run_load_file = self.init(**kwargs)  # 默认 False/None
            run_load_file = not run_load_file
        if hasattr(self, "load_file") and run_load_file:
            try:
                self.load_file()
            except Exception:
                traceback.print_exc()
        if self._check_filled and miss_list:
            raise OptionNotFilled(
                f"option: {miss_list} is not filled in {self.name}"
            )
        self.flush_option()

    def __str__(self):
        return (
            f"<{self.__class__.__name__} {self.name}>"
            if self.name
            else f"<{self.__class__.__name__}>"
        )

    def __repr__(self):
        return self.__str__()

    if TYPE_CHECKING:
        _options: Dict[str, Union[Callable, object]] = {}

        def init(self, **kwargs) -> bool:  # type: ignore
            """如果子类定义了这个函数，则会在 __init__ 之后调用这个函数
            返回值为 True 则不会调用 load_file 函数
            """

        def load_file(self) -> bool:  # type: ignore
            """如果子类定义了这个函数，则会在 __init__ 和 init 之后再调用这个函数

            请注意，这个函数请尽量使用 try 包裹住可能出现错误的部分
            否则会在控制台输出你的报错"""
            return True

    def option(self) -> Dict[str, Any]:
        """
        获取配置类的所有配置
        :return: 自己的所有配置
        """
        values = {}
        for ann in self.__annotations__:  # 获取类型注释
            values[ann] = getattr(self, ann, None)
            if values[ann] is None:
                values[ann] = self.__annotations__[ann]

        if not hasattr(self, "_options"):
            self._options: Dict[str, Union[Callable, object]] = {}
        for option, a_fun in self._options.items():  # 获取额外内容
            values[option] = a_fun

        for option, a_fun in values.items():  # 检查是否为 property
            if a_fun is bool and getattr(self, option, None) is not None:
                values[option] = False
            if isinstance(a_fun, property):
                try:
                    values[option] = getattr(self, option)
                except AttributeError:
                    raise OptionNotFound(
                        f"Option {option} is not found in {self.name}"
                    ) from None
        return values

    def str_option(
        self, shrink_to_long: Optional[int] = None
    ) -> Dict[str, Union[str, Any]]:
        """
        获取配置类的所有配置 并将所有非 BuiltIn 类型的值转换为 str
        :return:
        """
        raw_option = self.option()
        str_option = to_str_value_(raw_option)
        if shrink_to_long is None:
            return str_option
        if not isinstance(shrink_to_long, int) or shrink_to_long <= 0:
            return str_option
        for option, value in str_option.items():
            if value is not None:
                if len(str(value)) > shrink_to_long:
                    str_option[option] = str(value)[:shrink_to_long] + "..."
        return str_option

    def format(self, text: str) -> str:
        """
        使用自己的选项给输入的字符串替换内容
        :param text: 想替换的内容
        :return: 替换之后的内容
        """
        cache_option = self.flush_option()
        for option, value in cache_option.items():
            text = text.replace(f"{{{option}}}", str(value))
        return text

    def flush_option(self) -> Dict[str, Any]:
        """
        刷新缓存 options 的内容
        :return: 刷新过的 options
        """
        self.cached_options = self.option()
        return self.cached_options

    def option_with_len(
        self
    ) -> Tuple[List[Tuple[str, Any, Type]], int, int, int]:
        """
        返回一个可以用于打印的 option 列表
        :return:
        """
        options = self.flush_option()
        max_len_key = 1
        max_len_value = 1
        max_len_value_t = 1
        option_list = []
        for key, value in options.items():
            value_t = (
                type(value) if isinstance(value, type(value)) else type(value)
            )  # 判定这个类型 是不是 基本类型
            max_len_key = max(max_len_key, len(key))
            max_len_value = max(max_len_value, len(str(value)))
            max_len_value_t = max(max_len_value_t, len(str(value_t)))
            option_list.append([key, value, value_t])
        return [option_list, max_len_key, max_len_value, max_len_value_t]  # type: ignore

    def as_markdown(self, longest: Optional[int] = None) -> str:
        """
        返回一个 markdown 格式的 option 字符串
        :param longest: 最长的输出长度
        :return: markdown 格式的 option 字符串
        """
        value = self.option_with_len()
        cache = StringIO()
        option_len = max(value[1], len("Option"))
        value_len = max(value[2], len("Value"))
        value_type_len = max(value[3], len("Value Type"))

        # | Option | Value | Value Type |
        shortest = len("Option | Value | Value Type")

        if longest is not None:
            console_width = max(longest, shortest)
        else:
            console_width = shutil.get_terminal_size(fallback=(100, 80)).columns
            console_width = max(console_width, shortest)

        # 为每一栏 预分配 1/3 或者 需要的宽度 (如果不需要 1/3)
        option_len = min(option_len, console_width // 3)
        value_len = min(value_len, console_width // 3)
        value_type_len = min(value_type_len, console_width // 3)

        # 先指定每一个列的输出最窄宽度, 然后去尝试增加宽度
        # 循环分配新空间之前 首先检查是否已经不需要多分配 (and 后面)
        while option_len + value_len + value_type_len + 16 < console_width and (
            option_len < value[1]
            or value_len < value[2]
            or value_type_len < value[3]
        ):
            # 每一个部分的逻辑都是
            # 如果现在的输出长度小于原始长度
            # 并且长度 + 1 之后的总长度依然在允许范围内
            # 那么就 + 1
            if (
                option_len < value[1]
                and option_len + value_len + value_type_len + 16 < console_width
            ):
                option_len += 1
            if (
                value_len < value[2]
                and option_len + value_len + value_type_len + 16 < console_width
            ):
                value_len += 1
            if (
                value_type_len < value[3]
                and option_len + value_len + value_type_len + 16 < console_width
            ):
                value_type_len += 1
        # 实际上 对于列表(可变对象) for 出来的这个值是一个引用
        # 所以可以直接修改 string
        for v in value[0]:
            if len(str(v[0])) > option_len:
                v[0] = f"{str(v[0])[:value_len - 3]}..."  # type: ignore
            if len(str(v[1])) > value_len:
                v[1] = f"{str(v[1])[:value_len - 3]}..."  # type: ignore
            if len(str(v[2])) > value_type_len:
                v[2] = f"{str(v[2])[:value_len - 3]}.."  # type: ignore

        cache.write(
            f"| Option{' ' * (option_len - 3)}| Value{' ' * (value_len - 2)}"
            f"| Value Type{' ' * (value_type_len - 7)}|\n"
        )
        cache.write(
            f'|:{"-" * (option_len + 3)}|:{"-" * (value_len + 3)}|:{"-" * (value_type_len + 3)}|\n'
        )
        for option, value, value_t in value[0]:
            cache.write(
                f"| `{option}`{' ' * (option_len - len(option))} "
                f"| `{value}`{' ' * (value_len - len(str(value)))} "
                f"| `{value_t}`{' ' * (value_type_len - len(str(value_t)))} |\n"
            )
        result = cache.getvalue()
        cache.close()
        return result

    @classmethod
    def add_option(cls, name: str, value: Union[Callable, object]) -> Dict:
        """
        向配置类中添加一个额外的配置
        :param name: 配置的名字
        :param value: 用于获取配置的函数或者类
        :return: 配置类的所有配置
        """
        if not hasattr(cls, "_options"):
            cls._options: Dict[str, Union[Callable, object]] = {}
        cls._options[name] = value
        return cls._options

    @staticmethod
    def init_option(
        options_class: Type["Options"], init_value: Optional[dict] = None
    ) -> "Options":
        return options_class(**init_value if init_value is not None else {})



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

"""
TMD 啊啊啊啊啊啊啊啊啊
这里不是用来直接调用的啊啊啊啊啊
"""

from lib_not_dr.nuitka.reader import cli_main

if __name__ == "__main__":
    cli_main()

main = cli_main



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------
import sys

from typing import Dict, TYPE_CHECKING

from lib_not_dr.types.options import Options

COLOR_SUPPORT = True

if sys.platform == "win32":
    try:
        # https://stackoverflow.com/questions/36760127/...
        # how-to-use-the-new-support-for-ansi-escape-sequences-in-the-windows-10-console
        from ctypes import windll

        kernel32 = windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    except OSError:  # pragma: no cover
        COLOR_SUPPORT = False


class LogLevel(Options):
    name = "LogLevel"
    notset: int = 0
    notset_name: str = "NOTSET"
    trace: int = 5
    trace_name: str = "TRACE"
    fine: int = 7
    fine_name: str = "FINE"
    debug: int = 10
    debug_name: str = "DEBUG"
    info: int = 20
    info_name: str = "INFO"
    warn: int = 30
    warn_name: str = "WARN"
    error: int = 40
    error_name: str = "ERROR"
    fatal: int = 50
    fatal_name: str = "FATAL"

    level_name_map: Dict[int, str] = {
        notset: notset_name,
        trace:  trace_name,
        fine:   fine_name,
        debug:  debug_name,
        info:   info_name,
        warn:   warn_name,
        error:  error_name,
        fatal:  fatal_name,
    }

    name_level_map: Dict[str, int] = {
        notset_name: notset,
        trace_name:  trace,
        fine_name:   fine,
        debug_name:  debug,
        info_name:   info,
        warn_name:   warn,
        error_name:  error,
        fatal_name:  fatal,
    }

    @classmethod
    def parse_name_level(cls, name: str) -> int:
        """
        parse logging name to level int
        :param name: logging level name
        """
        name = name.upper()
        if name in cls.name_level_map:
            return cls.name_level_map[name]
        else:
            return cls.notset

    @classmethod
    def parse_level_name(cls, level_int: int, use_upper: bool = True) -> str:
        """
        parse logging level int to name
        :param level_int: logging level int
        :param use_upper:
        """
        if (result := cls.level_name_map.get(level_int)) is not None:
            return result

        if use_upper:
            for level in cls.level_name_map:
                if level_int <= level:
                    return cls.level_name_map[level]
        else:
            for level in reversed(cls.level_name_map):
                if level_int >= level:
                    return cls.level_name_map[level]
        return cls.notset_name


# fmt: off
from lib_not_dr.loggers.config import get_logger, get_config, read_config  # noqa: E402
from lib_not_dr.loggers import config  # noqa: E402

if TYPE_CHECKING:
    from lib_not_dr.loggers import logger
    from lib_not_dr.loggers import formatter
    from lib_not_dr.loggers import outstream
    from lib_not_dr.loggers import structure

__all__ = [
    # modules
    'logger',
    'formatter',
    'outstream',
    'structure',
    'config',
    # class
    'LogLevel',
    # function,
    'get_logger',
    'get_config',
    'read_config',
    # variable
    'COLOR_SUPPORT',
]
# fmt: on



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import time

from pathlib import Path
from types import FrameType
from typing import List, Optional, Tuple, Dict, Union

__all__ = ["LogMessage", "FormattingMessage"]


class LogMessage:
    # 消息内容本身的属性
    messages: List[str] = []
    end: str = "\n"
    split: str = " "

    # 消息的属性
    flush: Optional[bool] = None
    level: int = 20  # info
    log_time: float = 0.0  # time.time_ns() if None
    logger_name: str = "root"
    logger_tag: Optional[str] = None
    stack_trace: Optional[FrameType] = None

    def __init__(
        self,
        messages: Optional[List[str]] = None,
        end: str = "\n",
        split: str = " ",
        flush: Optional[bool] = None,
        level: int = 20,
        log_time: Optional[float] = None,
        logger_name: str = "root",
        logger_tag: Optional[str] = None,
        stack_trace: Optional[FrameType] = None,
    ) -> None:
        """
        Init for LogMessage
        :param messages: message list for log
        :param end: end of message
        :param split: split for messages
        :param flush: do flush or not
        :param level: level of message
        :param log_time: time of message (default: time.time_ns())
        :param logger_name: name of loggers
        :param logger_tag: tag of loggers
        :param stack_trace: stack trace of loggers
        """
        # 20231128 23:23
        # 因为 Options 的初始化太慢了 所以改为不继承 直接编写

        self.messages = messages if messages is not None else []
        self.end = end
        self.split = split
        self.level = level
        self.logger_name = logger_name
        self.logger_tag = logger_tag
        self.stack_trace = stack_trace

        if log_time is None:
            log_time = time.time_ns()
        self.log_time = log_time

        if flush is not None and not isinstance(flush, bool):
            flush = True if flush else False

        self.flush = flush

    def option(self) -> dict:
        return {
            "messages": self.messages,
            "end": self.end,
            "split": self.split,
            "flush": self.flush,
            "level": self.level,
            "log_time": self.log_time,
            "logger_name": self.logger_name,
            "logger_tag": self.logger_tag,
            "stack_trace": self.stack_trace,
        }

    def format_message(self) -> str:
        if self.split is None:
            self.split = " "
        return self.split.join(str(item) for item in self.messages) + self.end

    def format_for_message(self) -> Dict[str, str]:
        basic_info = self.option()

        if self.logger_tag is None:
            basic_info["logger_tag"] = "   "

        basic_info["messages"] = self.format_message()

        return basic_info

    @property
    def create_msec_3(self) -> int:
        return int(self.log_time / 1000000) % 1000


FormattingMessage = Tuple[LogMessage, Dict[str, Union[str, Path]]]



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import sys
import traceback

from pathlib import Path
from warnings import warn
from typing import Iterable, List

from lib_not_dr.nuitka import nuitka_config_type, raw_config_type, parse_config_function


def pyproject_toml(toml_data: dict) -> raw_config_type:
    """
    :param toml_data: dict (from pyproject/ raw dict)
    :return: dict
    """
    if "tool" not in toml_data:
        raise ValueError("No tool section in config file/dict")

    if "lndl" not in toml_data["tool"]:
        raise ValueError("No lib-not-dr(lndl) section in config file/dict")

    if "nuitka" not in toml_data["tool"]["lndl"]:
        raise ValueError(
            "No lib-not-dr(lndl).nuitka section in config file/dict"
        )

    nuitka_config = toml_data["tool"]["lndl"]["nuitka"]

    if "cli" not in nuitka_config:
        raise ValueError(
            "No lib-not-dr(lndl).nuitka.cli section in config file/dict"
        )

    if "main" not in nuitka_config["cli"]:
        raise ValueError(
            "'main' not define in lib-not-dr(lndl).nuitka.cli section\n"
            "define it with 'main = [<main.py>]'"
        )

    return nuitka_config


def toml_path_cli() -> Path:
    """
    get toml path from cli args
    :return: Path
    """
    if len(sys.argv) < 2:
        raw_path = Path().cwd()
    else:
        raw_path = Path(sys.argv[1])
    if raw_path.is_file():
        return raw_path

    elif raw_path.is_dir():
        if (raw_path / "pyproject.toml").exists():
            return raw_path / "pyproject.toml"
        else:
            raise FileNotFoundError(f"pyproject.toml not found in {raw_path}")
    else:
        raise FileNotFoundError(f"{raw_path} not found")


def get_cli_nuitka_args() -> dict:
    """
    get -- from sys.argv
    :return: list
    """
    # no -- in sys.argv
    if len(sys.argv) < 2:
        return {}
    if "--" not in sys.argv:
        return {}

    # start from --
    index = sys.argv.index("--")
    new_args = sys.argv[index + 1 :]
    arg_dict = {}
    for arg in new_args:
        if not arg.startswith("--"):
            warn(f"invalid arg: {arg}")
        else:
            arg = arg[2:]  # remove --
            # arg_name: --<name>=<value>
        arg_name = arg.split("=")[0]
        if "=" in arg:
            arg_value = arg.split("=")[1]
        else:
            arg_value = True
        arg_dict[arg_name] = arg_value

    return arg_dict


def merge_cli_config(toml_config: dict, cli_config: dict) -> dict:
    """
    merge toml config and cli config
    :param toml_config:
    :param cli_config:
    :return:
    """
    for name, value in cli_config.items():
        if name in toml_config:
            warn(
                "\033[33mcli config will overwrite toml config\n"
                f"{name}:{toml_config[name]} -> {value}\033[0m"
            )
            if isinstance(toml_config[name], bool):
                if not isinstance(value, bool):
                    warn(
                        f"cli config {name} is bool but toml config is not\n{value} -> {value}"
                    )
                    continue
                toml_config[name] = value
            elif isinstance(toml_config[name], str):
                if not isinstance(value, str):
                    warn(
                        f"cli config {name} is str but toml config is not\n{value} -> {value}"
                    )
                    continue
                toml_config[name] = value
            elif isinstance(toml_config[name], Iterable):
                if not isinstance(value, str):
                    warn(
                        f"cli config {name} is Iterable but toml config is not\n{value} -> {value}"
                    )
                    continue
                toml_config[name].append(value)
        else:
            toml_config[name] = value
    return toml_config


def gen_subprocess_args(
    nuitka_config: nuitka_config_type
) -> list:
    cmd_list = [sys.executable, "-m", "nuitka"]

    nuitka_config = merge_cli_config(nuitka_config, get_cli_nuitka_args())

    def parse_value(arg_name, arg_value) -> list:
        if isinstance(arg_value, bool):
            warn(f"bool value is not supported in list config {arg_name}")
            return []
        elif isinstance(arg_value, str):
            return [f"--{arg_name}={arg_value}"]
        else:
            return [f"--{arg_name}={arg_value[0]}={arg_value[1]}"]

    for name, value in nuitka_config.items():
        if isinstance(value, bool):
            # --<name>
            if value:
                cmd_list.append(f"--{name}")
            continue
        elif isinstance(value, str):
            # --<name>=<value>
            cmd_list.append(f"--{name}={value}")
            continue
        elif isinstance(value, Iterable):
            if "__spilt__" in value:
                # --<name>=<value1> --<name>=<value2> ...
                for item in value:
                    if item == "__spilt__":
                        continue
                    cmd_list += parse_value(name, item)
                continue
            if all(isinstance(item, str) for item in value):
                # --<name>=<value1>,<value2>,...
                cmd_list.append(f"--{name}={','.join(value)}")  # type: ignore
            elif all(isinstance(item, (tuple, list)) for item in value):
                # --<name>=<value1>=<value2>
                # --<name>=<value3>=<value4>,...
                for item in value:
                    cmd_list.append(f"--{name}={item[0]}={item[1]}")
            else:
                # 处理混杂的情况
                for item in value:
                    cmd_list += parse_value(name, item)
                continue
        warn(f"invalid config {name}:{value}")

    return cmd_list


def parse_raw_config_by_script(raw_config: raw_config_type) -> nuitka_config_type:
    """
    parse raw config by script
    整个函数就是一大堆检查加上一点点逻辑
    真就是逻辑 10 行 安全检查 100 行
    :param raw_config:
    :return: parsed config
    """
    raw_cli_config: nuitka_config_type = raw_config.get("cli", {})  # type: ignore

    if (script_name := raw_config.get("script")) is None:  # type: ignore
        return raw_cli_config
    script_name: str
    print(f'reading script {script_name}')
    script_path = Path(script_name)

    sys.path.append(str(script_path.parent))

    try:
        script_module = __import__(script_path.stem)
    except Exception as e:
        print(f"script {script_path} import failed ignore it\n{e}")
        sys.path.remove(str(script_path.parent))
        return raw_cli_config

    sys.path.remove(str(script_path.parent))

    if not hasattr(script_module, "main"):
        print(f"script {script_path} has no paser function ignore it")
        return raw_cli_config

    parse_func: parse_config_function = getattr(script_module, "main")
    try:
        parsed_config = parse_func(raw_config)
    except Exception as e:
        print(f"script {script_path} parse failed ignore it")
        print(e)
        traceback.print_exc()
        return raw_cli_config

    return parsed_config


def subprocess_to_bash(cmd_list: List[str]) -> str:
    """
    :param cmd_list: list
    :return: str
    """
    cmd_list = [item if " " not in item else f'"{item}"' for item in cmd_list]
    return " ".join(cmd_list)



from dataclasses import dataclass, field
from typing import Set, List


class Parsed:
    ...


@dataclass
class Option:
    name: str
    shortcuts: List[str]
    optional: bool
    types: Set[type] = field(default_factory=lambda: {str})


@dataclass
class OptionGroup:
    options: List[Option]
    optional: bool = True
    exclusive: bool = False


@dataclass
class Argument:
    name: str
    types: Set[type] = field(default_factory=lambda: {str})


@dataclass
class Flag:
    name: str
    shortcuts: List[str]


@dataclass
class FlagGroup:
    flags: List[Flag]
    exclusive: bool = False



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lib_not_dr import loggers, nuitka, types, command

_version_ = "0.3.15"

# fmt: off
__all__ = [
    "_version_",
    "loggers",
    "nuitka",
    "types",
    "command",
]
# fmt: on



# 本文件以 GNU Lesser General Public License v3.0（GNU LGPL v3) 开源协议进行授权 (谢谢狐狸写出这么好的MCDR)
# 顺便说一句,我把所有的tab都改成了空格,因为我觉得空格比tab更好看(草,后半句是github copilot自动填充的)

"""
This part of code come from MCDReforged(https://github.com/Fallen-Breath/MCDReforged)
Thanks a lot to Fallen_Breath and MCDR contributors
GNU Lesser General Public License v3.0 (GNU LGPL v3)
"""

import re
from typing import List, Callable, Tuple, Optional, Union

"""
Plugin Version
"""


# beta.3 -> (beta, 3), random -> (random, None)
class ExtraElement:
    DIVIDER = "."
    body: str
    num: Optional[int]

    def __init__(self, segment_str: str):
        segments = segment_str.rsplit(self.DIVIDER, 1)
        try:
            self.body, self.num = segments[0], int(segments[1])
        except (IndexError, ValueError):
            self.body, self.num = segment_str, None

    def __str__(self):
        if self.num is None:
            return self.body
        return "{}{}{}".format(self.body, self.DIVIDER, self.num)

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError()
        if self.num is None or other.num is None:
            return str(self) < str(other)
        else:
            return (self.body, self.num) < (other.body, other.num)


class Version:
    """
    A version container that stores semver like version string

    Example:

    * ``"1.2.3"``
    * ``"1.0.*"``
    * ``"1.2.3-pre4+build.5"``
    """

    EXTRA_ID_PATTERN = re.compile(r"|[-+0-9A-Za-z]+(\.[-+0-9A-Za-z]+)*")
    WILDCARDS = ("*", "x", "X")
    WILDCARD = -1

    component: List[int]
    has_wildcard: bool
    pre: Optional[ExtraElement]
    build: Optional[ExtraElement]

    def __init__(self, version_str: str, *, allow_wildcard: bool = True):
        """
        :param version_str: The version string to be parsed
        :keyword allow_wildcard: If wildcard (``"*"``, ``"x"``, ``"X"``) is allowed. Default: ``True``
        """
        if not isinstance(version_str, str):
            raise VersionParsingError("Invalid input version string")

        def separate_extra(text, char) -> Tuple[str, Optional[ExtraElement]]:
            if char in text:
                text, extra_str = text.split(char, 1)
                if not self.EXTRA_ID_PATTERN.fullmatch(extra_str):
                    raise VersionParsingError(
                        "Invalid build string: " + extra_str
                    )
                extra = ExtraElement(extra_str)
            else:
                extra = None
            return text, extra

        self.component = []
        self.has_wildcard = False
        version_str, self.build = separate_extra(version_str, "+")
        version_str, self.pre = separate_extra(version_str, "-")
        if len(version_str) == 0:
            raise VersionParsingError("Version string is empty")
        for comp in version_str.split("."):
            if comp in self.WILDCARDS:
                self.component.append(self.WILDCARD)
                self.has_wildcard = True
                if not allow_wildcard:
                    raise VersionParsingError(
                        "Wildcard {} is not allowed".format(comp)
                    )
            else:
                try:
                    num = int(comp)
                except ValueError:
                    num = None
                if num is None:
                    raise VersionParsingError(
                        "Invalid version number component: {}".format(comp)
                    )
                if num < 0:
                    raise VersionParsingError(
                        "Unsupported negatived number component: {}".format(num)
                    )
                self.component.append(num)
        if len(self.component) == 0:
            raise VersionParsingError("Empty version string")

    def __str__(self):
        version_str = ".".join(
            map(
                lambda c: str(c) if c != self.WILDCARD else self.WILDCARDS[0],
                self.component,
            )
        )
        if self.pre is not None:
            version_str += "-" + str(self.pre)
        if self.build is not None:
            version_str += "+" + str(self.build)
        return version_str

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, index: int) -> int:
        if index < len(self.component):
            return self.component[index]
        else:
            return (
                self.WILDCARD
                if self.component[len(self.component) - 1] == self.WILDCARD
                else 0
            )

    def __lt__(self, other):
        if not isinstance(other, Version):
            raise TypeError(
                "Cannot compare between instances of {} and {}".format(
                    Version.__name__, type(other).__name__
                )
            )
        for i in range(max(len(self.component), len(other.component))):
            if self[i] == self.WILDCARD or other[i] == self.WILDCARD:
                continue
            if self[i] != other[i]:
                return self[i] < other[i]
        if self.pre is not None and other.pre is not None:
            return self.pre < other.pre
        elif self.pre is not None:
            return not other.has_wildcard
        elif other.pre is not None:
            return False
        else:
            return False

    def __eq__(self, other):
        return not self < other and not other < self

    def __le__(self, other):
        return self == other or self < other

    def compare_to(self, other):
        if self < other:
            return -1
        elif self > other:
            return 1
        else:
            return 0


DEFAULT_CRITERION_OPERATOR = "="


class Criterion:
    def __init__(
        self,
        opt: str,
        base_version: Version,
        criterion: Callable[[Version, Version], bool],
    ):
        self.opt = opt
        self.base_version = base_version
        self.criterion = criterion

    def test(self, target: Union[Version, str]):
        return self.criterion(self.base_version, target)

    def __str__(self):
        return "{}{}".format(
            self.opt if self.opt != DEFAULT_CRITERION_OPERATOR else "",
            self.base_version,
        )


class VersionRequirement:
    """
    A version requirement tester

    It can test if a given :class:`Version` object matches its requirement
    """

    CRITERIONS = {
        "<=": lambda base, ver: ver <= base,
        ">=": lambda base, ver: ver >= base,
        "<": lambda base, ver: ver < base,
        ">": lambda base, ver: ver > base,
        "=": lambda base, ver: ver == base,
        "^": lambda base, ver: ver >= base and ver[0] == base[0],
        "~": lambda base, ver: ver >= base
        and ver[0] == base[0]
        and ver[1] == base[1],
    }

    def __init__(self, requirements: str):
        """
        :param requirements: The requirement string, which contains several version predicates connected by space character.
            e.g. ``">=1.0.x"``, ``"^2.9"``, ``">=1.2.0 <1.4.3"``
        """
        if not isinstance(requirements, str):
            raise VersionParsingError(
                "Requirements should be a str, not {}".format(
                    type(requirements).__name__
                )
            )
        self.criterions = []  # type: List[Criterion]
        for requirement in requirements.split(" "):
            if len(requirement) > 0:
                for prefix, func in self.CRITERIONS.items():
                    if requirement.startswith(prefix):
                        opt = prefix
                        base_version = requirement[len(prefix) :]
                        break
                else:
                    opt = DEFAULT_CRITERION_OPERATOR
                    base_version = requirement
                self.criterions.append(
                    Criterion(opt, Version(base_version), self.CRITERIONS[opt])
                )

    def accept(self, version: Union[Version, str]):
        if isinstance(version, str):
            version = Version(version)
        for criterion in self.criterions:
            if not criterion.test(version):
                return False
        return True

    def __str__(self):
        return " ".join(map(str, self.criterions))


class VersionParsingError(ValueError):
    pass



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from typing import List, Set, Dict, Optional, Tuple

from lib_not_dr.loggers.logger import Logger
from lib_not_dr.loggers.formatter import BaseFormatter
from lib_not_dr.loggers.outstream import BaseOutputStream
from lib_not_dr.loggers import formatter, outstream, LogLevel
from lib_not_dr.types.options import Options, OptionNameNotDefined


class ConfigStorage(Options):
    name = "LoggerConfigStorage"

    # 存储 logger, formatter, output 的字典
    loggers: Dict[str, Logger] = {}
    formatters: Dict[str, BaseFormatter] = {}
    outputs: Dict[str, BaseOutputStream] = {}
    # 存储失败的 logger, formatter, output 的字典
    fail_loggers: Dict[str, dict] = {}
    fail_formatters: Dict[str, dict] = {}
    fail_outputs: Dict[str, dict] = {}

    log: Logger = Logger(logger_name="loggers-storage")

    def have_formatter(self, formatter_name: str) -> bool:
        return formatter_name in self.formatters

    def have_output(self, output_name: str) -> bool:
        return output_name in self.outputs

    def have_logger(self, logger_name: str) -> bool:
        return logger_name in self.loggers

    def merge_storage(self, other_storage: "ConfigStorage") -> None:
        """
        Merge storage
        :param other_storage:
        :return:
        """
        self.loggers.update(other_storage.loggers)
        self.formatters.update(other_storage.formatters)
        self.outputs.update(other_storage.outputs)
        self.fail_loggers.update(other_storage.fail_loggers)
        self.fail_formatters.update(other_storage.fail_formatters)
        self.fail_outputs.update(other_storage.fail_outputs)

    # by GitHub Copilot
    @classmethod
    def _detect_cycle(
        cls,
        graph: Dict[str, List[str]],
        start: str,
        visited: Set[str],
        path: List[str],
    ) -> List[str]:
        visited.add(start)  # 将当前节点添加到已访问的节点集合中
        path.append(start)  # 将当前节点添加到当前路径中
        for neighbour in graph[start]:  # 遍历当前节点的所有邻居
            if (
                neighbour in visited
            ):  # 如果邻居节点已经被访问过，那么我们找到了一个循环
                return path + [neighbour]  # 返回包含循环的路径
            cycle_path = cls._detect_cycle(
                graph, neighbour, visited, path
            )  # 递归地在邻居节点上调用函数
            if cycle_path:  # 如果在邻居节点上找到了循环，那么返回包含循环的路径
                return cycle_path
        visited.remove(start)  # 从已访问的节点集合中移除当前节点
        path.remove(start)  # 从当前路径中移除当前节点
        return []  # 如果没有找到循环，那么返回一个空列表

    @classmethod
    def find_cycles(cls, graph: Dict[str, List[str]]) -> List[str]:
        cycles_set = set()  # 创建一个集合来存储所有的循环
        for node in graph:  # 遍历图中的所有节点
            cycle = cls._detect_cycle(
                graph, node, set(), []
            )  # 在每个节点上调用detect_cycle函数
            if cycle:  # 如果找到了循环，那么将循环添加到集合中
                cycles_set.update(cycle)
        return sorted(cycles_set)  # 返回排序后的循环列表

    def parse_level(self, level_config: dict) -> Optional[int]:
        """
        Parse level config
        :param level_config:
        :return:
        """
        level_found: Tuple[Optional[int], Optional[str]] = (
            level_config.get("level"),
            level_config.get("level_name"),
        )
        if all(_l is None for _l in level_found):
            # 如果都没有
            self.log.warn(
                f"No level or level_name in config {level_config}, ignored"
            )
            return 20
        if all(_l is not None for _l in level_found):
            # 如果都有 那么使用 level_name
            self.log.warn(
                f"Level and level_name both exist in config {level_config}, using level_name"
            )
            # 去掉 level 保留 level_name
            level_found = (
                None,
                level_found[1],
            )
        if level_found[0] is not None:
            # 如果 level 存在 那么使用 level
            return level_found[0]
        if level_found[1] is not None:
            # 如果 level_name 存在 那么使用 level_name
            return LogLevel.parse_name_level(level_found[1])
        return None

    def get_class_by_name(
        self, config: Dict[str, str], module
    ) -> Optional[type]:
        """
        Get class by name
        :param config:
        :param module:
        :return:
        """
        # check class
        if "class" not in config:
            self.log.warn(f"No class in config {config}, ignored")
            return None
        class_name = config.pop("class")
        # check class exist
        if class_name not in module.__all__:
            self.log.warn(
                f"Class {class_name} not found in module {module}, ignored"
            )
            return None

        return getattr(module, class_name)

    def parse_formatter(self, formatter_config: Dict[str, dict]) -> None:
        """
        Parse formatter config
        :param formatter_config:
        :return:
        """
        env = ConfigStorage()

        # Check circle require
        formatter_require = {}
        for key, value in formatter_config.items():
            if "sub_formatter" in value:
                formatter_require[key] = value["sub_formatter"]
            else:
                formatter_require[key] = []
        cycles_require = self.find_cycles(formatter_require)
        # 去除循环依赖
        if cycles_require:
            for formatter_name in cycles_require:
                self.log.error(
                    f"Formatter {formatter_name} have a cycle require, ignored"
                )
                env.fail_formatters[formatter_name] = formatter_config[
                    formatter_name
                ]
                formatter_config.pop(formatter_name)
        # Parse formatter
        ensure = 1000
        while len(formatter_config) > 0:
            pop_list = []
            for key, value in formatter_config.items():
                key: str
                value: dict
                formatter_name = key
                # check class
                formatter_class = self.get_class_by_name(value, formatter)
                if formatter_class is None:
                    env.fail_formatters[formatter_name] = value
                    pop_list.append(key)
                    continue
                # ensure sub formatter exist
                if "sub_formatter" in value:
                    fmts = []
                    for fmt in value["sub_formatter"]:
                        if not env.have_formatter(fmt):
                            formatter_config[formatter_name] = value
                            continue
                        fmts.append(env.formatters[fmt])
                    value["sub_formatter"] = fmts
                # init formatter
                try:
                    formatter_instance = formatter_class(**value)
                except OptionNameNotDefined as e:
                    self.log.error(
                        f"Formatter {formatter_name} class {formatter_class} init failed, ignored\n"
                        f"Error: {e}"
                    )
                    env.fail_formatters[formatter_name] = value
                    pop_list.append(key)
                    continue
                # add formatter
                env.formatters[formatter_name] = formatter_instance
                pop_list.append(key)
            # pop formatter
            for key in pop_list:
                formatter_config.pop(key)

            ensure -= 1
            if ensure <= 0:
                self.log.error(
                    "Formatter parse failed, ignored\n"
                    f"Left formatters: {formatter_config}"
                )
                # add all left formatter to fail formatter
                env.fail_formatters.update(formatter_config)
                break
        self.merge_storage(env)
        return None

    def parse_output(self, output_config: Dict[str, dict]) -> None:
        """
        Parse output config
        :param output_config:
        :return:
        """
        env = ConfigStorage()
        for output_name, config in output_config.items():
            # check class
            output_class = self.get_class_by_name(config, outstream)
            if output_class is None:
                env.fail_outputs[output_name] = config
                continue
            # get formatter for output
            if "formatter" in config:
                if self.formatters.get(config["formatter"]) is None:
                    if self.fail_formatters.get(config["formatter"]) is None:
                        self.log.error(
                            f'Output {output_name} formatter {config["formatter"]} not found, ignored'
                        )
                    else:
                        self.log.error(
                            f'Output {output_name} require a fail formatter {config["formatter"]}, ignored'
                        )
                    env.fail_outputs[output_name] = config
                    continue
                else:
                    config["formatter"] = self.formatters[config["formatter"]]
            if level := self.parse_level(config) is not None:
                config["level"] = level
            if "level_name" in config:
                config.pop("level_name")
            # init output
            try:
                output_instance = output_class(**config)
            except OptionNameNotDefined as e:
                self.log.error(
                    f"Output {output_name} class {output_class} init failed, ignored\n"
                    f"Error: {e}"
                )
                env.fail_outputs[output_name] = config
                continue
            # add output
            env.outputs[output_name] = output_instance
        self.merge_storage(env)
        return None

    def parse_logger(self, logger_config: Dict[str, dict]) -> None:
        """
        Parse loggers config
        :param logger_config: config of loggers
        """
        env = ConfigStorage()
        for logger_name, config in logger_config.items():
            # get output for logger
            if "outputs" in config:
                for i, output_name in enumerate(config["outputs"]):
                    if self.outputs.get(output_name) is None:
                        if self.fail_outputs.get(output_name) is None:
                            self.log.error(
                                f'Logger {logger_name} output {output_name} not found, ignored'
                            )
                        else:
                            self.log.error(
                                f'Logger {logger_name} require a fail output {output_name}, ignored'
                            )
                        env.fail_loggers[logger_name] = config
                        continue
                    else:
                        config["outputs"][i] = self.outputs[output_name]
            if (level := self.parse_level(config)) is not None:
                config["level"] = level
            if "level_name" in config:
                config.pop("level_name")
            # init logger
            try:
                logger_instance = Logger(**config)
            except OptionNameNotDefined as e:
                self.log.error(
                    f"Logger {logger_name} init failed, ignored\n" f"Error: {e}"
                )
                env.fail_loggers[logger_name] = config
                continue
            # add logger
            env.loggers[logger_name] = logger_instance
        self.merge_storage(env)
        return None

    def read_dict_config(self, config: Dict[str, dict]) -> None:
        """
        Read config from dict
        :param config:
        :return:
        """
        self.parse_formatter(config.get("Formatter", {}))
        self.parse_output(config.get("Outstream", {}))
        self.parse_logger(config.get("Logger", {}))


_storage = ConfigStorage(loggers={'root': Logger(logger_name='root')})


def get_config() -> ConfigStorage:
    global _storage
    return _storage


def get_logger(name: str = 'root', storage: Optional[ConfigStorage] = None) -> Logger:
    if storage is None:
        global _storage
        storage = _storage

    if name not in storage.loggers:
        root_log = storage.loggers['root'].clone_logger()
        root_log.logger_name = name
        storage.loggers[name] = root_log
    return storage.loggers[name]


def get_logger_from_old(name: str, old_name: str = 'root', storage: Optional[ConfigStorage] = None) -> Logger:
    if storage is None:
        global _storage
        storage = _storage

    if name not in storage.loggers:
        root_log = get_logger(old_name, storage).clone_logger()
        root_log.logger_name = name
        storage.loggers[name] = root_log
    return storage.loggers[name]


def add_logger(name: str, logger: Logger, storage: Optional[ConfigStorage] = None) -> None:
    if storage is None:
        global _storage
        storage = _storage

    storage.loggers[name] = logger


def read_config(log_config: Dict, storage: Optional[ConfigStorage] = None) -> ConfigStorage:
    if storage is None:
        global _storage
        storage = _storage

    storage.read_dict_config(log_config)
    return storage


# fmt: off
__all__ = [
    "get_config",
    "get_logger",
    "_storage",
    "read_config",
    "ConfigStorage",
]
# 整的跟 export 一样
# fmt: on



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------
import time

from pathlib import Path
from string import Template
from typing import List, Union, Optional, TYPE_CHECKING

from lib_not_dr.loggers import LogLevel
from lib_not_dr.types.options import Options
from lib_not_dr.loggers.structure import LogMessage, FormattingMessage

if TYPE_CHECKING:
    from lib_not_dr.loggers.formatter.colors import BaseColorFormatter

__all__ = ["BaseFormatter", "MainFormatter", "StdFormatter"]


class BaseFormatter(Options):
    name = "BaseFormatter"

    sub_formatter: List["BaseFormatter"] = []
    color_formatters: List["BaseColorFormatter"] = []
    default_template: str = (
        "[${log_time}][${level}]|${logger_name}:${logger_tag}|${messages}"
    )

    @classmethod
    def add_info(cls, match: str, to: str, description: str) -> str:
        return f"- {to} -> ${{{match}}} : {description}"

    @classmethod
    def info(cls) -> str:
        infos = {BaseFormatter.name: BaseFormatter._info()}
        cache = ""
        for formatter in cls.sub_formatter:
            infos[formatter.name] = formatter._info()
        infos[cls.name] = cls._info()
        for name, info in infos.items():
            cache += f"## {name}\n"
            cache += info
            cache += "\n"
        return cache

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info(
            "logger_name", "loggers name", "The name of the loggers"
        )
        info += "\n"
        info += cls.add_info(
            "logger_tag", "loggers tag", "The tag of the loggers"
        )
        return info

    def format_message(
        self,
        message: LogMessage,
        template: Optional[Union[Template, str]] = None,
    ) -> str:
        """
        Format message
        :param message: 输入的消息
        :param template: 日志输出模板
        :return:
        """
        basic_info = message.format_for_message()
        message, info = self._format((message, basic_info))

        if template is None:
            template = Template(self.default_template)
        elif isinstance(template, str):
            template = Template(template)

        try:
            return template.substitute(info)
        except (KeyError, ValueError):
            return template.safe_substitute(info)

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        """
        Format message
        :param message:
        :return:
        """
        for formatter in self.sub_formatter:
            message = formatter._format(message)
        return message

    @property
    def template(self) -> str:
        return self.default_template

    @template.setter
    def template(self, template: str) -> None:
        if not isinstance(template, str):
            raise TypeError(f"The template must be str, not {type(template)}")
        self.default_template = template


class MainFormatter(BaseFormatter):
    name = "TraceFormatter"
    sub_formatter: List["BaseFormatter"] = []

    time_format: str = "%Y-%m-%d %H:%M:%S"
    msec_time_format: str = "{}-{:03d}"
    use_absolute_path: bool = False

    default_level: int = 20
    level_get_higher: bool = True

    level_name_map = {
        LogLevel.notset: "NOTSET",
        LogLevel.trace: " TRACE",
        LogLevel.fine: " FINE ",
        LogLevel.debug: " DEBUG",
        LogLevel.info: " INFO ",
        LogLevel.warn: " WARN ",
        LogLevel.error: "ERROR ",
        LogLevel.fatal: "FATAL ",
    }
    name_level_map = {
        "NOTSET": LogLevel.notset,
        " TRACE": LogLevel.trace,
        " FINE ": LogLevel.fine,
        " DEBUG": LogLevel.debug,
        " INFO ": LogLevel.info,
        " WARN ": LogLevel.warn,
        "ERROR ": LogLevel.error,
        "FATAL ": LogLevel.fatal,
    }

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info(
            "log_time",
            "formatted time when logging",
            "The time format string"
            ". See https://docs.python.org/3/library/time"
            ".html#time.strftime for more information.",
        )
        info += "\n"
        info += cls.add_info("level", "log level", "The log level")
        info += "\n"
        info += cls.add_info(
            "log_source", "logging file", "the logging file name"
        )
        info += "\n"
        info += cls.add_info(
            "log_line", "logging line", "the logging line number"
        )
        info += "\n"
        info += cls.add_info(
            "log_function", "logging function", "the logging function name"
        )
        return info

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        message = self._time_format(message)
        message = self._trace_format(message)
        message = self._level_format(message)
        return message

    def _level_format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].level in self.name_level_map:
            level_tag = self.level_name_map[message[0].level]
        else:
            if self.level_get_higher:
                for level in self.name_level_map:
                    if message[0].level <= self.name_level_map[level]:
                        level_tag = level
                        break
                else:
                    level_tag = "FATAL"
            else:
                for level in self.name_level_map:
                    if message[0].level >= self.name_level_map[level]:
                        level_tag = level
                        break
                else:
                    level_tag = "NOTSET"
        message[1]["level"] = level_tag
        return message

    def _time_format(self, message: FormattingMessage) -> FormattingMessage:
        time_mark = time.localtime(message[0].log_time / 1000000000)
        if self.msec_time_format:
            time_mark = self.msec_time_format.format(
                time.strftime(self.time_format, time_mark),
                message[0].create_msec_3,
            )
        message[1]["log_time"] = time_mark
        return message

    def _trace_format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].stack_trace is None:
            return message
        raw_path = message[0].stack_trace.f_code.co_filename
        if self.use_absolute_path:
            path = Path(raw_path)
            message[1]["log_source"] = path.absolute()
        else:
            message[1]["log_source"] = raw_path
        message[1]["log_line"] = message[0].stack_trace.f_lineno
        message[1]["log_function"] = message[0].stack_trace.f_code.co_name
        return message


class StdFormatter(BaseFormatter):
    name = "StdFormatter"

    enable_color: bool = True

    sub_formatter: List[BaseFormatter] = [MainFormatter()]
    from lib_not_dr.loggers.formatter.colors import (
        LevelColorFormatter,
        LoggerColorFormatter,
        TimeColorFormatter,
        TraceColorFormatter,
        MessageColorFormatter,
    )

    color_formatters: List[BaseFormatter] = [
        LevelColorFormatter(),
        LoggerColorFormatter(),
        TimeColorFormatter(),
        TraceColorFormatter(),
        MessageColorFormatter(),
    ]

    default_template: str = (
        "[${log_time}][${level}]|${logger_name}:${logger_tag}|${messages}"
    )

    def __init__(
        self,
        enable_color: bool = True,
        sub_formatter: Optional[List[BaseFormatter]] = None,
        main_formatter: Optional[MainFormatter] = None,
        color_formatters: Optional[List[BaseFormatter]] = None,
        default_template: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initialize the StdFormatter
        :param enable_color: enable color
        :param sub_formatter: list of sub formatter
        :param main_formatter: main formatter
        :param color_formatters: list of color formatter
        :param default_template: default template
        :param kwargs: other options
        """
        # 同 structures.LogMessage.__init__ 的注释 (逃)
        self.enable_color = enable_color
        if sub_formatter is not None:
            self.sub_formatter = sub_formatter
        if color_formatters is not None:
            self.color_formatters = color_formatters
        if main_formatter is not None and isinstance(
            main_formatter, MainFormatter
        ):
            self.main_formatter = main_formatter
        else:
            self.main_formatter = MainFormatter()
        if default_template is not None:
            self.default_template = default_template
        super().__init__(**kwargs)

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        super()._format(message)

        if not self.enable_color:
            return message

        for formatter in self.color_formatters:
            message = formatter._format(message)

        return message

    @classmethod
    def _info(cls) -> str:
        return "None"


if __name__ == "__main__":
    import inspect

    log_message = LogMessage(
        messages=["Hello World!"],
        level=7,
        stack_trace=inspect.currentframe(),
        logger_tag="tester",
        logger_name="test",
    )

    print(StdFormatter.info())
    print(StdFormatter().format_message(log_message))

    std_format = StdFormatter()
    std_format.default_template = "${log_time}|${logger_name}|${logger_tag}|${log_source}:${log_line}|${log_function}|${level}|${messages}"

    test_levels = (0, 5, 7, 10, 20, 30, 40, 50)

    print("with color")

    for test_level in test_levels:
        log_message.level = test_level
        print(std_format.format_message(log_message), end="")

    print("without color")

    std_format.enable_color = False

    for test_level in test_levels:
        log_message.level = test_level
        print(std_format.format_message(log_message), end="")

    print(std_format.as_markdown())



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------
import sys
import time
import platform
import subprocess

from pathlib import Path

from lib_not_dr.nuitka import nuitka_config_type

from lib_not_dr.nuitka.reader.arg_parser import (
    pyproject_toml,
    toml_path_cli,
    gen_subprocess_args,
    subprocess_to_bash,
    parse_raw_config_by_script,
)

#  it will
# - read the given file or
# - find pyproject.toml in the given dir or
# - find pyproject.toml in current dir

# and read the `[tool.lndl.nuitka]` section
# then it will run nuitka with the given config

USEAGE = """
usage:
python lndl-nuitka.py --help (or -h)
show this help message

python lndl-nuitka.py <path-to-pyproject.toml>
python lndl-nuitka.py <path-to-dir>
python lndl-nuitka.py (with nothing, will use current dir)

arguments:
--no-run (-n) : only show the config, not run nuitka
--yes (-y) : run nuitka without asking
-- --xxx : pass to nuitka

用法:
python lndl-nuitka.py --help (或 -h)
显示这个帮助信息

python lndl-nuitka.py <一个文件>
python lndl-nuitka.py <一个路径>
python lndl-nuitka.py (直接运行 会使用当前目录)

参数:
--no-run (-n) : 只显示配置, 不运行 nuitka
--yes (-y) : 不询问直接运行 nuitka
-- --xxx : 传递给 nuitka 的参数

"""

TOML_READERS = (
    # "tomllib",  # stdlib toml reader after python 3.11
    # try tomllib first
    "toml",  # slow pure python toml reader
    "rtoml",  # rust based toml reader
    "tomlkit",  # pure python toml reader
    "tomli",  # pure python toml reader
    "pytomlpp",  # cpp based toml reader
    "qtoml",  # pure python toml reader, but faster than toml
)


def get_toml_reader():
    # 3.11 + 使用 tomllib
    if sys.version_info >= (3, 11):
        from tomllib import loads

        return loads
    for module_name in TOML_READERS:
        try:
            loaded = __import__(module_name).loads
            return loaded
        except ImportError:
            continue
    error_msg = """\033[31mNo toml reader found, please install any below by pip:
    {toml} or use Python 3.11+ or install lib-not-dr with lib_not_dr[nuitka] option
    没有找到 toml 解析器, 请考虑安装下列之一的 toml 库
    {toml} 或者使用 Python 3.11+ 或者使用 lib_not_dr[nuitka] 安装 lib-not-dr\033[0m
    """.format(toml="\n    ".join(TOML_READERS))
    raise ImportError(error_msg) from None


toml_loads = get_toml_reader()


def display_config(subprocess_command: list) -> None:
    print(f"The config is:\n\033[34m{subprocess_command} \033[0m")
    print("shell command is:\n\033[34m", end="")
    print(subprocess_to_bash(subprocess_command), "\033[0m")
    print(f"Working Dir: \033[32m {Path().cwd().absolute()} \033[0m")


def run_nuitka(subprocess_command: list) -> None:
    start_time = time.time()
    # shell true on windows
    # shell false on linux
    if platform.system() == "Windows":
        subprocess.run(subprocess_command, shell=True, check=True)
    else:
        subprocess.run(subprocess_command, shell=False, check=True)
    end_time = time.time()
    print(f"Time Elapsed: {end_time - start_time} seconds")


def main(config: nuitka_config_type) -> list:
    """
    enter point for python direct call
    :param config: nuitka config dict
    :return: None
    """
    subprocess_command = gen_subprocess_args(config)
    display_config(subprocess_command)
    return subprocess_command


def cli_main() -> None:
    """
    entering point of cli
    :return: None
    """
    if "--help" in sys.argv or "-h" in sys.argv:
        print(USEAGE)
        sys.exit(0)

    if len(sys.argv) < 2:
        print(USEAGE)
        if input("are you sure to run? (y/n)") not in ["y", "Y", "yes", "Yes"]:
            sys.exit(0)

    toml_file = toml_path_cli()

    with open(toml_file, "r", encoding="utf-8") as f:
        toml = toml_loads(f.read())

    nuitka_config = pyproject_toml(toml)
    cli_config = parse_raw_config_by_script(nuitka_config)

    subprocess_command = gen_subprocess_args(cli_config)
    display_config(subprocess_command)

    exit_arg = ("-no-run", "-n")
    confirm_arg = ("-y", "-yes")
    if any(arg in sys.argv for arg in exit_arg):
        sys.exit(0)
    elif all(arg not in sys.argv for arg in confirm_arg):
        if input("are you sure to run? (y/n)") not in ["y", "Y", "yes", "Yes"]:
            sys.exit(0)

    run_nuitka(subprocess_command)



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from .options import (
    Options,
    OptionsError,
    OptionNotFound,
    OptionNameNotDefined,
    get_type_hints_,
)

from .version import (
    Version,
    VersionRequirement,
    VersionParsingError,
    ExtraElement,
)

__all__ = [
    # options
    "get_type_hints_",
    "Options",
    "OptionsError",
    "OptionNotFound",
    "OptionNameNotDefined",
    # version
    "Version",
    "VersionRequirement",
    "VersionParsingError",
    "ExtraElement",
]



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from lib_not_dr.loggers import LogLevel, COLOR_SUPPORT
from lib_not_dr.loggers.formatter import BaseFormatter
from lib_not_dr.loggers.structure import FormattingMessage

__all__ = [
    "BaseColorFormatter",
    "LevelColorFormatter",
    "LoggerColorFormatter",
    "TimeColorFormatter",
    "TraceColorFormatter",
    "MessageColorFormatter",
    "RESET_COLOR",
]

RESET_COLOR = "\033[0m"


class BaseColorFormatter(BaseFormatter):
    name = "BaseColorFormatter"
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: "",
        # Trace: blue
        LogLevel.trace: "\033[38;2;138;173;244m",
        # Fine: green
        LogLevel.fine: "\033[0;32m",
        # Debug: cyan
        LogLevel.debug: "\033[0;36m",
        # Info: white
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[0;33m",
        # Error: red
        LogLevel.error: "\033[0;31m",
        # Fatal: red background
        LogLevel.fatal: "\033[0;41m",
    }

    def get_color(self, message: FormattingMessage) -> str:
        if message[0].level in self.color:
            return self.color[message[0].level]
        for level in self.color:
            if message[0].level <= level:
                break
        else:
            level = 90
        return self.color[level]


class LevelColorFormatter(BaseColorFormatter):
    name = "LevelColorFormatter"
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: "",
        # Trace: blue
        LogLevel.trace: "\033[38;2;138;173;244m",
        # Fine: green
        LogLevel.fine: "\033[35;48;2;44;44;54m",
        # Debug: cyan
        LogLevel.debug: "\033[38;2;133;138;149m",
        # Info: white
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[0;33m",
        # Error: red
        LogLevel.error: "\033[0;31m",
        # Fatal: red background
        LogLevel.fatal: "\033[0;41m",
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info("colored level", "level", "A colored level")

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if isinstance(message[1].get("level"), int) or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == "" or color == RESET_COLOR:
            return message
        if "level" in message[1]:
            message[1]["level"] = f'{color}{message[1]["level"]}{RESET_COLOR}'
        return message


class LoggerColorFormatter(BaseColorFormatter):
    name = "LoggerColorFormatter"
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: "",
        # Trace: blue
        LogLevel.trace: "\033[38;2;138;173;244m",
        # Fine: green
        LogLevel.fine: "\033[0;32m",
        # Debug: cyan
        LogLevel.debug: "\033[0;36m",
        # Info: white
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[0;33m",
        # Error: red
        LogLevel.error: "\033[0;31m",
        # Fatal: red background
        LogLevel.fatal: "\033[38;2;245;189;230m",
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info(
            "colored loggers name", "loggers name", "A colored loggers name"
        )

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get("logger_name") is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == "" or color == RESET_COLOR:
            return message
        if "logger_name" in message[1]:
            message[1][
                "logger_name"
            ] = f'{color}{message[1]["logger_name"]}{RESET_COLOR}'
        if "logger_tag" in message[1] and message[1].get("logger_tag") != "   ":
            message[1][
                "logger_tag"
            ] = f'{color}{message[1]["logger_tag"]}{RESET_COLOR}'
        return message


class TimeColorFormatter(BaseColorFormatter):
    name = "TimeColorFormatter"
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: "",
        # Trace: blue
        LogLevel.trace: "\033[38;2;138;173;244m",
        # Fine: green
        LogLevel.fine: "\033[0;32m",
        # Debug: cyan
        LogLevel.debug: "\033[0;36m",
        # Info: white
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[0;33m",
        # Error: red
        LogLevel.error: "\033[0;31m",
        # Fatal: red background
        LogLevel.fatal: "\033[38;2;255;255;0;48;2;120;10;10m",
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info("colored time", "time", "A colored time")

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get("log_time") is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == "" or color == RESET_COLOR:
            return message
        if "log_time" in message[1]:
            message[1][
                "log_time"
            ] = f'{color}{message[1]["log_time"]}{RESET_COLOR}'
        return message


class TraceColorFormatter(BaseColorFormatter):
    name = "TraceColorFormatter"
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: "\033[38;2;0;255;180m",
        # Trace: blue
        LogLevel.trace: "\033[38;2;0;255;180m",
        # Fine: green
        LogLevel.fine: "\033[38;2;0;255;180m",
        # Debug: cyan
        LogLevel.debug: "\033[38;2;0;255;180m",
        # Info: white
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[38;2;0;255;180m",
        # Error: red
        LogLevel.error: "\033[38;2;0;255;180m",
        # Fatal: red background
        LogLevel.fatal: "\033[38;2;255;255;0;48;2;120;10;10m",
    }

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info(
            "colored logging file", "log_source", "A colored logging file name"
        )
        info += "\n"
        info += cls.add_info(
            "colored logging line", "log_line", "A colored logging line number"
        )
        info += "\n"
        info += cls.add_info(
            "colored logging function",
            "log_function",
            "A colored logging function name",
        )
        return info

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].stack_trace is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == "" or color == RESET_COLOR:
            return message
        for name in ("log_source", "log_line", "log_function"):
            if name in message[1]:
                message[1][name] = f"{color}{message[1][name]}{RESET_COLOR}"
        return message


class MessageColorFormatter(BaseColorFormatter):
    name = "MessageColorFormatter"

    color = {
        # Notset: just black
        LogLevel.notset: "",
        # Trace: blue
        LogLevel.trace: "\033[38;2;138;173;244m",
        # Fine: blue
        LogLevel.fine: "\033[38;2;138;173;244m",
        # Debug: blue
        LogLevel.debug: "\033[38;2;138;173;244m",
        # Info: no color
        LogLevel.info: "\033[0;38;2;255;255;255m",
        # Warn: yellow
        LogLevel.warn: "\033[0;33m",
        # Error: red
        LogLevel.error: "\033[0;31m",
        # Fatal: red background
        LogLevel.fatal: "\033[38;2;255;255;0;48;2;120;10;10m",
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info("colored message", "message", "A colored message")

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get("messages") is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == "" or color == RESET_COLOR:
            return message
        if "messages" in message[1]:
            if message[1]["messages"][-1] == "\n":
                message[1][
                    "messages"
                ] = f'{color}{message[1]["messages"][:-1]}{RESET_COLOR}\n'
            else:
                message[1][
                    "messages"
                ] = f'{color}{message[1]["messages"]}{RESET_COLOR}'
        return message



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import platform
import warnings
from pathlib import Path
from typing import List, Tuple, Optional, Union, Any
from enum import Enum

from lib_not_dr.types import Options, Version, VersionRequirement


def ensure_cmd_readable(cmd: str) -> str:
    """
    保证 参数中 不含空格
    :param cmd: 要格式化的命令行参数
    :return: 格式化后的命令行参数
    """
    if " " in str(cmd):
        return f'"{cmd}"'
    return cmd


def format_cmd(
    arg_name: Optional[str] = None,
    arg_value: Optional[Union[str, List[str]]] = None,
    write: Optional[Any] = True,
) -> List[str]:
    """
    用来格式化输出命令行参数
    :param arg_name: 类似 --show-memory 之类的主项
    :param arg_value: 类似 xxx 类的内容
    :param write: 是否写入
    :return: 直接拼接好的命令行参数 不带 =
    """
    if not write:
        return []
    if arg_name is None:
        return []
    if arg_value is None:
        return [arg_name]
    if isinstance(arg_value, list):
        arg_value = ",".join(
            [ensure_cmd_readable(value) for value in arg_value]
        )
        return [f"{arg_name}{arg_value}"]
    arg_value = ensure_cmd_readable(arg_value)
    return [f"{arg_name}{arg_value}"]


class NuitkaSubConfig(Options):
    """
    Nuitka 配置的子项
    Nuitka configuration sub-items
    """

    name = "Nuitka Sub Configuration"

    def gen_cmd(self) -> List[str]:
        """
        生成命令行参数
        :return:
        """
        raise NotImplementedError


class NuitkaPluginConfig(NuitkaSubConfig):
    """
    控制 nuitka 的 plugin 相关参数的部分
    Control part of nuitka's plugin related parameters
    """

    name = "Nuitka Plugin Configuration"

    # --enable-plugin=PLUGIN_NAME
    enable_plugin: List[str] = []
    # --disable-plugin=PLUGIN_NAME
    disable_plugin: List[str] = []
    # --plugin-no-detection
    plugin_no_detection: bool = False
    # --user-plugin=PATH
    user_plugin: List[Path] = []
    # --show-source-changes
    show_source_changes: bool = False

    # --include-plugin-directory=MODULE/PACKAGE
    include_plugin_dir: List[str] = []
    # --include-plugin-files=PATTERN
    include_plugin_files: List[str] = []

    def gen_cmd(self) -> List[str]:
        lst = []
        lst += format_cmd(
            "--enable-plugin=", self.enable_plugin, self.enable_plugin
        )
        lst += format_cmd(
            "--disable-plugin=", self.disable_plugin, self.disable_plugin
        )
        lst += format_cmd(
            "--plugin-no-detection" if self.plugin_no_detection else None
        )
        lst += format_cmd(
            "--user-plugin=",
            [str(plugin.absolute()) for plugin in self.user_plugin],
            self.user_plugin,
        )
        lst += format_cmd(
            "--show-source-changes" if self.show_source_changes else None
        )
        lst += format_cmd(
            "--include-plugin-directory=",
            self.include_plugin_dir,
            self.include_plugin_dir,
        )
        lst += format_cmd(
            "--include-plugin-files=",
            self.include_plugin_files,
            self.include_plugin_files,
        )
        return lst


class NuitkaIncludeConfig(NuitkaSubConfig):
    """
    控制 nuitka 的 include 和 数据 相关参数的部分
    Control part of nuitka's include related parameters
    """

    name = "Nuitka Include Configuration"

    # --include-package=PACKAGE
    include_packages: List[str] = []
    # --include-module=MODULE
    include_modules: List[str] = []

    # --prefer-source-code
    # --no-prefer-source-code    for --module
    prefer_source_code: bool = False
    # --follow-stdlib
    follow_stdlib: bool = False

    def gen_cmd(self) -> List[str]:
        lst = []
        lst += format_cmd(
            "--include-package=", self.include_packages, self.include_packages
        )
        lst += format_cmd(
            "--include-module=", self.include_modules, self.include_modules
        )
        lst += format_cmd(
            "--prefer-source-code" if self.prefer_source_code else None
        )
        lst += format_cmd(
            "--no-prefer-source-code" if not self.prefer_source_code else None
        )
        lst += format_cmd("--follow-stdlib" if self.follow_stdlib else None)
        return lst


class NuitkaDataConfig(NuitkaSubConfig):
    """
    控制 nuitka 的 数据 相关参数的部分
    Control part of nuitka's data related parameters
    """

    name = "Nuitka Data Configuration"

    # --include-package-data=PACKAGE=PACKAGE_PATH
    include_package_data: List[Tuple[Path, Path]] = []
    # --include-data-files=PATH=PATH
    include_data_files: List[Tuple[Path, Path]] = []
    # --include-data-dir=DIRECTORY=PATH
    include_data_dir: List[Tuple[Path, Path]] = []

    # --noinclude-data-files=PATH
    no_include_data_files: List[Path] = []

    # --list-package-data=LIST_PACKAGE_DATA
    list_package_data: List[str] = []
    # --list-package-dlls=LIST_PACKAGE_DLLS
    list_package_dlls: List[str] = []

    # --include-distribution-metadata=DISTRIBUTION
    include_distribution_metadata: List[str] = []


class NuitkaBinaryInfo(Options):
    """
    nuitka 构建的二进制文件的信息
    nuitka build binary file information
    """

    name = "Nuitka Binary Info"

    # --company-name=COMPANY_NAME
    company_name: Optional[str] = None
    # --product-name=PRODUCT_NAME
    product_name: Optional[str] = None

    # --file-version=FILE_VERSION
    # --macos-app-version=MACOS_APP_VERSION
    file_version: Optional[Union[str, Version]] = None
    # --product-version=PRODUCT_VERSION
    product_version: Optional[Union[str, Version]] = None

    # --file-description=FILE_DESCRIPTION
    file_description: Optional[str] = None
    # --copyright=COPYRIGHT_TEXT
    copyright: Optional[str] = None
    # --trademarks=TRADEMARK_TEXT
    trademarks: Optional[str] = None

    # Icon
    # --linux-icon=ICON_PATH
    # --macos-app-icon=ICON_PATH
    # --windows-icon-from-ico=ICON_PATH
    # --windows-icon-from-exe=ICON_EXE_PATH
    # 注意: 只有 Windows 下 才可以提供多个 ICO 文件
    # 其他平台 和 EXE 下只会使用第一个路径
    icon: Optional[List[Path]] = None

    # Console
    # --enable-console
    # --disable-console
    console: bool = True

    # Windows UAC
    # --windows-uac-admin
    windows_uac_admin: bool = False
    # --windows-uac-uiaccess
    windows_uac_ui_access: bool = False


class NuitkaOutputConfig(Options):
    """
    nuitka 构建的选项
    nuitka build output information
    """

    name = "Nuitka Output Config"

    # --output-dir=DIRECTORY
    output_dir: Optional[Path] = None
    # --output-filename=FILENAME
    output_filename: Optional[str] = None

    # --quiet
    quiet: bool = False
    # --no-progressbar
    no_progressbar: bool = False
    # --verbose
    verbose: bool = False
    # --verbose-output=PATH
    verbose_output: Optional[Path] = None

    # --show-progress
    show_progress: bool = False
    # --show-memory
    show_memory: bool = False
    # --show-scons
    show_scons: bool = False
    # --show-modules
    show_modules: bool = False
    # --show-modules-output=PATH
    show_modules_output: Optional[Path] = None

    # --xml=XML_FILENAME
    xml: Optional[Path] = None
    # --report=REPORT_FILENAME
    report: Optional[Path] = None
    # --report-diffable
    report_diffable: bool = False

    # --remove-output
    remove_output: bool = False
    # --no-pyo-file
    no_pyo_file: bool = False


class NuitkaDebugConfig(Options):
    """
    nuitka 构建的调试选项
    nuikta build debug information
    """

    name = "Nuitka Debug Config"

    # --debug
    debug: bool = False
    # --unstripped
    strip: bool = True
    # --profile
    profile: bool = False
    # --internal-graph
    internal_graph: bool = False
    # --trace-execution
    trace_execution: bool = False
    # --recompile-c-only
    recompile_c_only: bool = False
    # --generate-c-only
    generate_c_only: bool = False
    # --deployment
    deployment: bool = False
    # --no-deployment-flag=FLAG
    deployment_flag: Optional[str] = None
    # --experimental=FLAG
    experimental: Optional[str] = None


class NuitkaTarget(Enum):
    """
    用于指定 nuitka 构建的目标
    Use to specify the target of nuitka build
    exe: 不带任何参数
    module: --module
    standalone: --standalone
    one_file: --onefile
    """

    exe = ""
    module = "module"
    standalone = "standalone"
    one_file = "package"


class NuitkaScriptGenerator(Options):
    """
    用于帮助生成 nuitka 构建脚本的类
    Use to help generate nuitka build script

    :arg main 需要编译的文件
    """

    name = "Nuitka Script Generator"

    # --main=PATH
    # 可以有多个 输入时需要包在列表里
    main: List[Path]

    # --run
    run_after_build: bool = False
    # --debugger
    debugger: bool = False
    # --execute-with-pythonpath
    execute_with_python_path: bool = False

    # --assume-yes-for-downloads
    download_confirm: bool = True

    # standalone/one_file/module/exe
    target: NuitkaTarget = NuitkaTarget.exe

    # --python-debug
    python_debug: bool = False
    # --python-flag=FLAG
    python_flag: List[str] = []
    # --python-for-scons=PATH
    python_for_scons: Optional[Path] = None


class CompilerHelper(Options):
    """
    用于帮助生成 nuitka 构建脚本的类
    Use to help generate nuitka build script

    """

    name = "Nuitka Compiler Helper"

    output_path: Path = Path("./build")
    src_file: Path

    python_cmd: str = "python"
    compat_nuitka_version: VersionRequirement = VersionRequirement(
        "~1.8.0"
    )  # STATIC VERSION

    # 以下为 nuitka 的参数
    # nuitka options below
    use_lto: bool = False  # --lto=yes (no is faster)
    use_clang: bool = True  # --clang
    use_msvc: bool = True  # --msvc=latest
    use_mingw: bool = False  # --mingw64

    onefile: bool = False  # --onefile
    onefile_tempdir: Optional[str] = ""  # --onefile-tempdir-spec=
    standalone: bool = True  # --standalone
    use_ccache: bool = True  # not --disable-ccache
    enable_console: bool = True  # --enable-console / --disable-console

    show_progress: bool = True  # --show-progress
    show_memory: bool = False  # --show-memory
    remove_output: bool = True  # --remove-output
    save_xml: bool = False  # --xml
    xml_path: Path = Path("build/compile_data.xml")
    save_report: bool = False  # --report
    report_path: Path = Path("build/compile_report.xml")

    download_confirm: bool = True  # --assume-yes-for-download
    run_after_build: bool = False  # --run

    company_name: Optional[str] = ""
    product_name: Optional[str] = ""
    file_version: Optional[Version] = None
    product_version: Optional[Version] = None
    file_description: Optional[str] = ""  # --file-description

    copy_right: Optional[str] = ""  # --copyright

    icon_path: Optional[Path] = None

    follow_import: List[str] = []
    no_follow_import: List[str] = []

    include_data_dir: List[Tuple[str, str]] = []
    include_packages: List[str] = []

    enable_plugin: List[str] = []  # --enable-plugin=xxx,xxx
    disable_plugin: List[str] = []  # --disable-plugin=xxx,xxx

    def init(self, **kwargs) -> None:
        if (compat_version := kwargs.get("compat_nuitka_version")) is not None:
            if not self.compat_nuitka_version.accept(compat_version):
                warnings.warn(
                    f"Nuitka version may not compat with {compat_version}\n"
                    "requirement: {self.compat_nuitka_version}"
                )
        # 非 windows 平台不使用 msvc
        if platform.system() != "Windows":
            self.use_msvc = False
            self.use_mingw = False
        else:
            self.use_mingw = self.use_mingw and not self.use_msvc
            # Windows 平台下使用 msvc 时不使用 mingw

    def __str__(self):
        return self.as_markdown()

    def as_markdown(self, longest: Optional[int] = None) -> str:
        """
        输出编译器帮助信息
        Output compiler help information

        Args:
            longest (Optional[int], optional):
                输出信息的最大长度限制 The maximum length of output information.
                Defaults to None.

        Returns:
            str: 以 markdown 格式输出的编译器帮助信息
                Compile helper information in markdown format
        """
        front = super().as_markdown(longest)
        gen_cmd = self.gen_subprocess_cmd()
        return f"{front}\n\n```bash\n{' '.join(gen_cmd)}\n```"

    def gen_subprocess_cmd(self) -> List[str]:
        """生成 nuitka 构建脚本
        Generate nuitka build script

        Returns:
            List[str]:
                生成的 nuitka 构建脚本
                Generated nuitka build script
        """
        warnings.warn(
            "\033[33mlib_not_dr.nuitka.compile is deprecated "
            "use lib_not_dr.nuitka.reader.main or lndl-nuitka instead\033[0m",
            DeprecationWarning,
            stacklevel=2,
        )
        cmd_list = [self.python_cmd, "-m", "nuitka"]
        # macos 和 非 macos icon 参数不同
        if platform.system() == "Darwin":
            cmd_list += format_cmd(
                "--macos-app-version=",
                self.product_version,
                self.product_version,
            )
            if self.icon_path is not None:
                cmd_list += format_cmd(
                    "--macos-app-icon=", self.icon_path.absolute(), self.icon_path
                )
        elif platform.system() == "Windows" and self.icon_path is not None:
            cmd_list += format_cmd(
                "--windows-icon-from-ico=",
                self.icon_path.absolute(),
                self.icon_path,
            )
        elif platform.system() == "Linux" and self.icon_path is not None:
            cmd_list += format_cmd(
                "--linux-icon=", self.icon_path.absolute(), self.icon_path
            )

        cmd_list += format_cmd("--lto=", "yes" if self.use_lto else "no")
        cmd_list += format_cmd("--clang" if self.use_clang else None)
        cmd_list += format_cmd("--msvc=latest" if self.use_msvc else None)
        cmd_list += format_cmd("--mingw64" if self.use_mingw else None)
        cmd_list += format_cmd("--standalone" if self.standalone else None)
        cmd_list += format_cmd("--onefile" if self.onefile else None)
        cmd_list += format_cmd(
            "--onefile-tempdir-spec=",
            self.onefile_tempdir,
            self.onefile_tempdir,
        )

        cmd_list += format_cmd(
            "--disable-ccache" if not self.use_ccache else None
        )
        cmd_list += format_cmd(
            "--show-progress" if self.show_progress else None
        )
        cmd_list += format_cmd("--show-memory" if self.show_memory else None)
        cmd_list += format_cmd(
            "--remove-output" if self.remove_output else None
        )
        cmd_list += format_cmd(
            "--assume-yes-for-download" if self.download_confirm else None
        )
        cmd_list += format_cmd("--run" if self.run_after_build else None)
        cmd_list += format_cmd(
            "--enable-console" if self.enable_console else "--disable-console"
        )

        cmd_list += format_cmd(
            "--xml=", str(self.xml_path.absolute()), self.save_xml
        )
        cmd_list += format_cmd(
            "--report=", str(self.report_path.absolute()), self.save_report
        )
        cmd_list += format_cmd(
            "--output-dir=", str(self.output_path.absolute()), self.output_path
        )
        cmd_list += format_cmd(
            "--company-name=", self.company_name, self.company_name
        )
        cmd_list += format_cmd(
            "--product-name=", self.product_name, self.product_name
        )
        cmd_list += format_cmd(
            "--file-version=", str(self.file_version), self.file_version
        )
        cmd_list += format_cmd(
            "--product-version=",
            str(self.product_version),
            self.product_version,
        )
        cmd_list += format_cmd(
            "--file-description=", self.file_description, self.file_description
        )
        cmd_list += format_cmd("--copyright=", self.copy_right, self.copy_right)

        cmd_list += format_cmd(
            "--follow-import-to=", self.follow_import, self.follow_import
        )
        cmd_list += format_cmd(
            "--nofollow-import-to=",
            self.no_follow_import,
            self.no_follow_import,
        )
        cmd_list += format_cmd(
            "--enable-plugin=", self.enable_plugin, self.enable_plugin
        )
        cmd_list += format_cmd(
            "--disable-plugin=", self.disable_plugin, self.disable_plugin
        )

        if self.include_data_dir:
            cmd_list += [
                f"--include-data-dir={src}={dst}"
                for src, dst in self.include_data_dir
            ]
        if self.include_packages:
            cmd_list += [
                f"--include-package={package}"
                for package in self.include_packages
            ]

        cmd_list.append(f"--main={self.src_file}")
        return cmd_list



class CallBackDescriptor:
    def __init__(self, name):
        self.callback_name = name

    def __set__(self, instance, value):
        assert (
            getattr(instance, self.callback_name) is None
        ), f"Attribute '{self.callback_name}' has been set."
        instance.__dict__[self.callback_name] = value

    def __get__(self, instance, owner):
        return (
            self
            if instance is None
            else instance.__dict__.get(self.callback_name)
        )



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import re
from typing import Callable, List, Optional, Union, Set

from .data import Option, Argument, Flag, Parsed
from .descriptor import CallBackDescriptor

try:
    from typing import Self
except ImportError:
    from typing import TypeVar

    Self = TypeVar("Self")  # NOQA

from .exception import IllegalName

CallBack = Union[
    Callable[[str], None], str
]  # Equals to `Callable[[str], None] | str`
# 可调用对象或字符串作为回调
# A callable or str as callback

ParseArgFunc = Callable[[str], Optional[type]]
# 解析参数的函数，返回值为 None 时表示解析失败
# function to parse argument, return None when failed

EMPTY_WORDS = re.compile(r"\s", re.I)


def check_name(name: Union[str, List[str]]) -> None:
    """
    Check the name or shortcuts of argument(s) or flag(s).
    The name must not be empty str, and must not contains \\t or \\n or \\f or \\r.
    If that not satisfy the requirements, it will raise exception `IllegalArgumentName`.
    检查 参数或标记 的 名称或快捷方式 是否符合要求。
    名称必须是非空的字符串，且不能包含 \\t 或 \\n 或 \\f 或 \\r。
    如果不符合要求，将会抛出 `IllegalArgumentName` 异常。
    :param name: arguments
    :return: None
    """
    if isinstance(name, str) and EMPTY_WORDS.search(name):
        raise IllegalName("The name of argument must not contains empty words.")
    elif isinstance(name, list) and all(
        (not isinstance(i, str)) and EMPTY_WORDS.search(i) for i in name
    ):
        raise IllegalName(
            "The name of shortcut must be 'str', and must not contains empty words."
        )
    else:
        raise TypeError("The type of name must be 'str' or 'list[str]'.")


class Literal:
    _tip = CallBackDescriptor("_tip")
    _func = CallBackDescriptor("_func")
    _err_callback = CallBackDescriptor("_err_callback")

    def __init__(self, name: str):
        self.name: str = name
        self.sub: List[Self] = []
        self._tip: Optional[CallBack] = None
        self._func: Optional[CallBack] = None
        self._err_callback: Optional[CallBack] = None

        self._opts: List[Option] = []
        self._args: List[Argument] = []
        self._flags: List[Flag] = []

    def __call__(self, *nodes) -> Self:
        self.sub += nodes
        return self

    def __repr__(self):
        attrs = (
            k
            for k in self.__dict__
            if not (k.startswith("__") and k.endswith("__"))
        )
        return f"{self.__class__.__name__}({', '.join(f'{k}={v!r}' for k in attrs if (v := self.__dict__[k]))})"

    def arg(self, name: str, types: Optional[Set[type]] = None) -> Self:
        Argument(name=name, types=types)
        return self

    def opt(
        self,
        name: str,
        shortcuts: Optional[List[str]] = None,
        optional: bool = True,
        types: Optional[Set[type]] = None,
    ) -> Self:
        check_name(name)
        if shortcuts is not None and len(shortcuts) != 0:
            check_name(shortcuts)
        self._opts.append(
            Option(
                name=name, shortcuts=shortcuts, optional=optional, types=types
            )
        )
        return self

    def opt_group(self, opts: List[Option], exclusive: bool = False):
        ...

    def flag(self, name: str, shortcuts: Optional[List[str]] = None) -> Self:
        check_name(name)
        if shortcuts is not None and len(shortcuts) != 0:
            check_name(shortcuts)
        Flag(name=name, shortcuts=shortcuts)
        ...
        return self

    def flag_group(self, flags: List[Flag], exclusive: bool = False) -> Self:
        ...
        return self

    def error(self, callback: CallBack) -> Self:
        self._err_callback = callback
        return self

    def run(self, func: CallBack) -> Self:
        self._func = func
        return self

    def tip(self, tip: CallBack) -> Self:
        self._tip = tip
        return self

    def parse(self, cmd: Union[str, List[str]]) -> Parsed:
        ...

    def to_doc(self) -> str:
        ...


def builder(node: Literal) -> Literal:
    ...
